//
// Author:    Jae Chung
// Date:      10/20/99
// File Name: cbt.h
//

#ifndef ns_cbt_h
#define ns_cbt_h

#include "queue.h"

class LinkDelay;


// RED early drop parameters, supplied by user
struct edp {
        // User supplied.
        int mean_pktsize;	// avg pkt size, linked into Tcl
	int bytes;		// true if queue in bytes, false if packets
	int wait;		// true for waiting between dropped packets
	int setbit;		// true to set congestion indication bit

	double th_min;		// minimum threshold of average queue size
	double th_max;		// maximum threshold of average queue size
	double max_p_inv;	// 1/max_p, for max_p = maximum prob.
	double q_w;		// queue weight given to cur q size sample

	// Computed.
        double ptc;		// packet time constant in packets/second
};


// RED early drop variables, maintained by RED
struct edv {
        TracedDouble v_ave;	// average queue size
        TracedDouble v_prob1;	// prob. of packet drop before "count".
        double v_slope;		// used in computing average queue size
	double v_prob;		// prob. of packet drop
	double v_a;		// v_prob = v_a * v_ave + v_b
	double v_b;
        int count;		// # of packets since last drop
        int count_bytes;	// # of bytes since last drop
        int old;		// 0 when average queue first exceeds thresh

	edv() : v_ave(0.0), v_prob1(0.0), v_slope(0.0), v_prob(0.0),
		v_a(0.0), v_b(0.0), count(0), count_bytes(0), old(0) { }
};

// CBT parameters
struct cbt {
	double th_max;             // maximum threshold
	int cq_size;            // class_queue's current size (logical queue)
        int cq_size_byte;       // class_queue's current size (in bytes) 

	cbt() : cq_size(0), cq_size_byte(0) { }
};


// CBT Queue Definitions
class CBTQueue : public Queue {
 public:
	CBTQueue();
 protected:
	int command(int argc, const char*const* argv);
	void enque(Packet* pkt);
	virtual Packet *pickPacketForECN(Packet* pkt);
	virtual Packet *pickPacketToDrop();
	Packet* deque();
	void reset();
	void run_estimator(int nqueued, int m);
	int drop_early(Packet* pkt);

	void calc_mm_q_avg(int m);
	void calc_udp_q_avg(int m);
	void cbt_accnt_enqueue(Packet* pkt);
	void cbt_accnt_dequeue(Packet* pkt);


	LinkDelay* link_;	// outgoing link
	int fifo_;		// fifo queue?
        PacketQueue *q_; 	// underlying (usually) FIFO queue
		
	int bcount_;	        // byte count
	int qib_;	        // bool: queue measured in bytes?
	NsObject* de_drop_;	// drop_early target

	Tcl_Channel tchan_;	// place to write trace records
	TracedInt curq_;	// current qlen seen by arrivals
	void trace(TracedVar*);	// routine to write trace records

	// Static state.
	int drop_tail_;		// drop-tail
	int drop_front_;	// drop-from-front
	int drop_rand_;		// drop-tail, or drop random?

	edp edp_;	// early-drop params
	int doubleq_;	// for experiments with priority for small packets
	int dqthresh_;	// for experiments with priority for small packets

	// Dynamic state.
	int idle_;		// queue is idle?
	double idletime_;	// if so, since this time
	edv edv_;		// early-drop variables

	// CBT state.
	cbt udp_class_;         // udp class
	cbt mmu_class_;         // mmu class
	int bypass_red_;        // bypass_red (udp,mmudp) on RED congestion

	int mm_idle_;           // mm queue is idle?
	double mm_idletime_;    // if so, since this time
	double mm_ptc_;         // mm packet time constant in packets/second
	double mm_q_avg_;       // mm queue average;
	double mm_q_weight_;    // mm queue weight

	int udp_idle_;           // udp queue is idle?
	double udp_idletime_;    // if so, since this time
	double udp_ptc_;         // udp packet time constant in packets/second
	double udp_q_avg_;       // udp queue average;
	double udp_q_weight_;    // mm queue weight


	double link_band_;       // output link bandwidth
	int off_ip_;            // ip header offset

	// For debugging
	void print_edp();
	void print_edv();
	void print_cbt();
};

#endif


